/*
 * Copyright (c) 2004 Center for Comparative and International Studies
 * at the Swiss Federal Institute of Technology Zurich
 * (see http://www.cis.ethz.ch/ for details)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Please see http://www.gnu.org/copyleft/gpl.txt for the full license text.
 *
 */
package ch.ethz.icr.geosim0;

import uchicago.src.sim.gui.Drawable;
import uchicago.src.sim.gui.SimGraphics;
import uchicago.src.sim.util.Random;

import java.awt.*;
import java.util.*;
import java.util.List;

/**
 * Geosim: Geopolitical Simulation Model
 * A model to simulate state formation with endogenous boundaries
 *
 * This model is based on the architecture introduced in Cederman, EMERGENT ACTORS IN WORLD
 * POLITICS. Princeton University Press, 1997. The current model corresponds to the more
 * recent article Cederman, "Modeling the Size of Wars: From Billiard Balls to Sandpiles,"
 * AMERICAN POLITICAL SCIENCE REVIEW 97 (February 2003): 135-150.
 *
 * The Actor class describes the behavior of the state actors that live in the
 * territorial grid. All of their functionality that's specific to individual
 * actor is specified here. Methods that link many actors together are usually
 * defined in Model.
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @version 3.0
 */
final class Actor implements Drawable {
    /**
     * Resource unit default size
     */
    private final double resUnit = 10.0;

    /**
     * Link to capital state
     */
    Actor capital;

    /**
     * Neighborhood list
     */
    final LinkedList neighs;

    /**
     * Identity number
     */
    final int id;

    /**
     * Position on the grid x-axis
     */
    final int x;

    /**
     * Position on the grid y-axis
     */
    final int y;

    /**
     * Resource level
     */
    double res;

    /**
     * Number of political neighbors
     */
    int numPolNeighs;         

    /**
     * Number of dependent provinces
     */
    int numProvinces;

    /**
     * Flag used in Model.topology()
     */
    boolean reached;

    /**
     * Most threatening state against which to balance
     */
    private Actor mainThreat;

    /**
     * A pointer to the model itself
     */
    private final Model model;

    /**
     * Flag used in initModule() (see Model)
     */
    boolean founder;

    /**
     * Constructor for actors (only called once since actors are indestructable)
     *
     * @param id
     * @param x
     * @param y
     * @param model
     */
    public Actor(final int id, final int x, final int y, final Model model) {

        this.model = model;

        this.id = id;
        this.x = x;
        this.y = y;

        // The initial power level is set to one resource unit except for a minority of the
        // cases where the power is multiplied by 100. This gives those states a headstart.
        // After the first time period, resources become entirely territorial, however.
        res = resUnit;

        numProvinces = 0;
        numPolNeighs = 0;
        capital = null;              // assume all initial actors are sovereign
        neighs = new LinkedList();  // create the neighborlist holding Relation objects

        mainThreat = null;

        founder = false;
    }

    final List getNeighbors() {
        return model.world.getVonNeumannNeighbors(x, y, false);
    }

    /**
     * Returns true if the unit is sovereign
     * @return
     */
    final boolean isSovereign() {
        return (capital == null);
    }


    /**
     * Returns true if the unit is an "atom",
     * i.e. a sovereign unitary state without provinces
     * @return
     */
    final boolean isAtom() {
        return (capital == null) && (numProvinces == 0);
    }

    /**
     * Returns true if the two units are (dependent) provinces in the same compound state
     * @param a
     * @return
     */
    private boolean isCompatriotProvince(final Actor a) {
        return ((capital != null) && (capital == a.capital));
    }

    /**
     * Returns true if the two units belong to the same compound actor
     * @param a
     * @return
     */
    final boolean isCompatriot(final Actor a) {
        return ((this == a) ||
                isCompatriotProvince(a) ||
                (capital == a) ||
                (a.capital == this));
    }

    /**
     * Returns true if the two units do not belong to the same compound actor
     * @param a
     * @return
     */
    final boolean isForeigner(final Actor a) {
        return !isCompatriot(a);
    }

    /**
     * Returns the number of political neighbors
     * this method is used to find the number of relations to loop through
     * because political neighbors are always put at the beginning of the list
     * NOTE: the secession option is not used here
     * @return
     */
    public final int getNumPoliticalNeighbors() {
        return numPolNeighs;
    }

    /**
     * Returns true if the actor, seen as a primitive unit,
     * is directly contiguious with any part of state a
     * NOTE: The argument must be a capital
     *
     * @param a
     * @return
     */
    private boolean isUnitAdjacentTo(final Actor a) {	// Unit means a basic territorial unit here
        boolean adjacent = false;
        final List neighbors = getNeighbors();
        for (Iterator iterator = neighbors.iterator(); iterator.hasNext() && !adjacent;) {
            final Actor neighbor = (Actor) iterator.next();
            adjacent = neighbor.isCompatriot(a);
        }
        return adjacent;
    }

    /**
     * Returns true if any part of the actor is continguious with any part of state a
     * NOTE: The argument must be a capital
     *
     * @param a
     * @return
     */
    final boolean isStateAdjacentTo(final Actor a) {
        boolean adjacent = isUnitAdjacentTo(a);
        if (!adjacent) {
            final ListIterator lit = neighs.listIterator();
            while ((lit.hasNext()) && (!adjacent)) {
                final Relation r = (Relation) lit.next();
                if (r.prov) {
                    adjacent = r.other.isUnitAdjacentTo(a);
                }
            }
        }
        return adjacent;
    }

    /**
     * Returns the capital province of the actor (is the same as the actor if atom)
     *
     * @return
     */
    final Actor getGovt() {
        if (isSovereign())
            return this;
        else
            return capital;
    }

    /**
     * Returns the Relation object with actor a taken from neighbor list
     *
     * @param a
     * @param list
     * @return
     */
    static final Relation getRelationWith(final Actor a, final List list) {
        final ListIterator lit = list.listIterator();
        while (lit.hasNext()) {
            final Relation r = (Relation) lit.next();
            if (r.other == a)
                return r;
        }
        return null;
    }

    /**
     * Resource updating
     *
     * Method that calculates the resource level of sovereign actors based on territorial size
     * Should be called only by sovereign states
     * Dependent provinces are also allocated resources through the taxation mechanism
     */
    final void updateResources() {
        double harvest = resUnit;  // start adding up resources, beginning with capital

        final ListIterator lit = neighs.listIterator();  // go through all dependent provinces
        for (int i = 0; i < neighs.size(); i++) {
            final Relation r = (Relation) lit.next();
            if (r.prov) {   // select out the provinces
                harvest = harvest + resUnit;   // the capital amassed tax revenue
            } else if (r.pol) { // add up damage for all political fronts (again, this is not used)
                r.oldRes = r.res;
            }

        }

        res = harvest;
    }

    /**
     * Alignment Phase
     *
     * Method that identifies the main threat facing the state
     */
    public final void perceive() {
        Actor threat = null;
        double minTrust = 0;     // trust scores are negative
        double trust;
        final ListIterator lit = neighs.listIterator();
        for (int i = 0; i < numPolNeighs; i++) {
            final Relation r = (Relation) lit.next();
            if (r.pol) {
                if (res <= 0.0)
                    trust = -r.twin.res;
                else // trust score is set to the negative local power balance
                    trust = -r.twin.res / r.res;
                if (trust < minTrust) { // local for the most threatening relationship
                    threat = r.other;
                    minTrust = trust;
                }
            }
        }

        if (minTrust <= model.minThreat) { // select main threat only if threat falls below minThreat
            mainThreat = threat;        // set the mainThreat indicator
        } else {
            mainThreat = null;          // otherwise, there's no threatening state to align against
            model.numSecureActors = model.numSecureActors + 1 + numProvinces;
        }
    }

    /**
     * Resource allocation
     *
     * Method that allocates the state's resources to the political relationships
     * both as fixed share and in proportion to the political neighbors' power
     * The propMobileResources variable determines the share of the resources that will be
     * allocated in a power-dependent way
     * The mobile (or fungible) resources are allocated under the assumption that
     * there be only one new conflict
     * This algorithm is a modified version of Proportional Resource Allocation
     * as described in Cederman, EMERGENT ACTORS, Chap. 5.
     */
    final void allocateResources() {
        double counterRes = 0.0;
        final double mobilRes = model.propMobileResources * res; // total share of fungible resources
        double fixedRes = 0.0;
        int numFronts = 0;

        ListIterator lit = neighs.listIterator();
        for (int i = 0; i < getNumPoliticalNeighbors(); i++) {  // go through all fronts looking for active ones
            final Relation r = (Relation) lit.next();
            numFronts++;
            if ((r.history[Relation.SELF] == Relation.D) ||
                    (r.history[Relation.OTHER] == Relation.D)) {
                counterRes = counterRes + r.twin.oldRes;
            }
        }

        if (numFronts != 0) // calculate the fixed resource share by dividing by the # of neighbors
            fixedRes = (res - mobilRes) / (double) numFronts;

        // at this point the state knows that it is facing counterRes res. units fighting

        // the state will now allocate resources to each front assuming that only one
        // new conflict would happen
        lit = neighs.listIterator();
        for (int i = 0; i < getNumPoliticalNeighbors(); i++) {
            final Relation r = (Relation) lit.next();
            if (counterRes == 0.0)
                r.res = fixedRes + mobilRes;
            else if ((r.history[Relation.SELF] == Relation.D) || // if the front is active then
                    (r.history[Relation.OTHER] == Relation.D))   // allocate fixed resources and
                r.res = fixedRes + (r.twin.oldRes / counterRes) * mobilRes; // proportional share
            else // otherwise, make the counterfactual assumption that the front had been active
                r.res = fixedRes + (r.twin.oldRes / (counterRes + r.twin.oldRes)) * mobilRes;
        }
    }

    /**
     * Method that selects a random contiguous target province inside
     * the political neighbor indicated by the relationship argument
     * A target province is the part that would be claimed in case of battle victory
     *
     * @param r
     */
    private void selectTarget(final Relation r) {
        if (r.other.numProvinces == 0) // if neighbor is atom then it will be the target
            r.target = r.other;
        else { // otherwise, we're dealing with a compound opponent
            final ArrayList candidates = new ArrayList();  // the target will be selected from this list
            if (r.other.isUnitAdjacentTo(this))  // start by adding capital if it's continguous
                candidates.add(r.other);
            final ListIterator lit = r.other.neighs.listIterator(); // then loop through all provinces
            for (int i = 0; i < r.other.neighs.size(); i++) {
                final Relation neigh = (Relation) lit.next();
                if (neigh.prov && (neigh.other.isUnitAdjacentTo(this)))
                    candidates.add(neigh.other);  // add the province if it is continguous
            }

            r.target = (Actor) candidates.get(// select a random target province from list
                    Random.uniform.nextIntFromTo(0, candidates.size() - 1));
        }
    }


    /**
     * Method that selects an agent province from where the attack would be launched
     * The agent province needs to be adjacent to the target province
     * @param r
     */
    private void selectAgent(final Relation r) {
        if (isAtom()) // if atom, then no need to select
            r.agent = this;
        else { // otherwise, the agent province can be in one of the four von-Neumann neighbors
            final ArrayList candidates = new ArrayList();   // of the target province
            final List neighbors = getNeighbors();
            for (Iterator iterator = neighbors.iterator(); iterator.hasNext();) {
                final Actor neighbor = (Actor) iterator.next();
                if ((neighbor != null) && (neighbor.isCompatriot(this))) {
                    candidates.add(neighbor);   // if the neighboring unit belongs to self then add to list
                }
            }
            r.agent = (Actor) candidates.get(// select a random member of the list
                    Random.uniform.nextIntFromTo(0, candidates.size() - 1));
        }
    }


    /**
     * Method that establishes a randomly selected "battle path" from the attacking state
     * to the attacked one
     * The agent province belongs to self and the target to the opponent
     * The target and agent provinces are stored inside the Relation object r
     * @param r
     */
    final void selectPath(final Relation r) {
        selectTarget(r);
        selectAgent(r);
    }

    /**
     * Decision phase
     *
     * Main decision-making method followed by all states
     * This routine stores the result in "act" field of the relational portfolios
     * Double buffering is used in by calling the "hist" field
     */
    final void decide() {
        int attacks = 0;        // a counter for number of active fronts
        final boolean tryAttack = Utility.bern(model.probAttack);
        ArrayList victims = null;  // the list of random targets of attacks

        if (tryAttack) {  // if an unprovoked attack will be attempted, we need the lists
            victims = new ArrayList();
        }

        // first play "grim trigger" with all political neighbors
        final ListIterator lit = neighs.listIterator();  // loop through the neighbors
        for (int i = 0; i < getNumPoliticalNeighbors(); i++) {  // we're only looking at political neighbors
            final Relation r = (Relation) lit.next(); // (the first relations are always political ones)
            r.act = r.history[Relation.OTHER];       // this is the tit-for-tat part of the strategy
            if (r.history[Relation.SELF] == Relation.D) // stick to defection according to grim trigger
                r.act = Relation.D;
            if (r.act == Relation.D)          // increment the attack counter for each defection
                attacks++;
            if (tryAttack &&
                    !r.prov &&
                    attacks == 0) {   // start process potential targets
                // continue to process targets of attacks, now randomly
                victims.add(r);
            }
        }

        // try an unprovoked attack
        if (tryAttack)
            if (attacks == 0) {  // don't attack if there are already active fronts
                final Relation r;
                if (victims.size() > 0) { // target a randomly chosen neighbor
                    r = (Relation) victims.get(Random.uniform.nextIntFromTo(0, victims.size() - 1));
                    selectPath(r);
                } else  // otherwise there is nobody to attack
                    r = null;
                if (r != null) {  // unless there is nobody to attack
                    if (strongerThan(r)) {   // if the power balance is favorable
                        r.act = Relation.D;        // carry out the attack
                        attacks++;
                    }
                }
            }
    }

    /**
     * Method returning the local power balance indicated by relation r
     *
     * @param r
     * @return
     */
    private static double powerRatio(final Relation r) {
        final double ownRes;
        final double otherRes;
        ownRes = r.res;         // start with the own relational resources
        otherRes = r.twin.res;  // and those of the opponent

        if (otherRes <= 0.0) // check if opponent's resources are negative or zero
            return Double.MAX_VALUE;
        else  // otherwise, return the ratio of own resources to other's resources
            return ownRes / otherRes;
    }

    /**
     * Method indicating if the state is stronger than state pointed to by relation r
     *
     * @param r
     * @return
     */
    private boolean strongerThan(final Relation r) {
        if (model.superioritySlope == 0)  // if deterministic criterion, just use the power ratio (above)
            return powerRatio(r) > model.superiorityThreshold;
        else  // if stochastic criterion, use logistic function
            return Utility.bern(Utility.logistic(powerRatio(r), model.superiorityThreshold, model.superioritySlope));
    }

    public final int getX() {
        return x;
    }

    public final int getY() {
        return y;
    }

    /**
     * Drawing method that draws the cell that belongs to the primitive actor
     * This method draws boundaries and capitals and marks war if the war counter is on
     *
     * @param g
     */
    public final void draw(final SimGraphics g) {
        final int wScale = g.getCellWidthScale();
        final int hScale = g.getCellHeightScale();

        final int scrX = wScale * getX();
        final int scrY = hScale * getY();
        g.setDrawingParameters(wScale, hScale, 0);

        if (getGovt().mainThreat != null)
            g.drawFastRect(Color.red);
          else
            g.drawFastRect(Color.white);

        if (getX() > 0 || getY() > 0) {

            Color color = Color.black;

            // Compute the thickness
            int width = (int) Math.floor(wScale / 10.0);
            int height = (int) Math.floor(hScale / 10.0);

            // Keep the thickness square
            if (width > height) {
                width = height;
            } else if (height > width) {
                height = width;
            }

            if (width > 0 && height > 0) {
                // Draw the first border
                if (getX() > 0) {
                    final Actor actor = (Actor) model.world.getObjectAt(getX() - 1, getY());
                    if (isForeigner(actor)) {
                        g.setDrawingParameters(width, hScale, 0);
                        g.drawFastRect(color);
                    }
                }

                // Draw the second border
                color = Color.black;
                if (getY() > 0) {
                    final Actor actor = (Actor) model.world.getObjectAt(getX(), getY() - 1);
                    if (isForeigner(actor)) {
                        g.setDrawingParameters(wScale, height, 0);
                        g.drawFastRect(color);
                    }
                }
            }
        }

        g.getGraphics().setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        if (isSovereign()) {
            final int divider = 2;
            final int additive = 0;
            // Compute the size of the dot
            int width = wScale / divider + additive; // 8
            int height = hScale / divider + additive; // 8

            // Keep it square
            if (width > height) {
                width = height;
            } else if (height > width) {
                height = width;
            }

            // Center the coordinate system
            g.setDrawingCoordinates(((scrX + wScale / 2) - width / 2) - additive,
                    ((scrY + hScale / 2) - height / 2) - additive,
                    0);

            g.setDrawingParametersNoScale(width, height, 0);

            if (isSovereign())
                g.drawFastCircle(Color.black);
            else {
                g.drawFastCircle(Color.red);
            }
        }
        g.getGraphics().setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_DEFAULT);
    }

}
