/*
 * Copyright (c) 2004 Center for Comparative and International Studies
 * at the Swiss Federal Institute of Technology Zurich
 * (see http://www.cis.ethz.ch/ for details)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Please see http://www.gnu.org/copyleft/gpl.txt for the full license text.
 *
 */
package ch.ethz.icr.geosim0;

import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.SimpleModel;
import uchicago.src.sim.space.Object2DGrid;
import uchicago.src.sim.util.Random;
import uchicago.src.sim.util.SimUtilities;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/**
 * Geosim: Geopolitical Simulation Model
 * A model to simulate state formation with endogenous boundaries
 *
 * This model is based on the architecture introduced in Cederman, EMERGENT ACTORS IN WORLD
 * POLITICS. Princeton University Press, 1997. The current model corresponds to the more
 * recent article Cederman, "Modeling the Size of Wars: From Billiard Balls to Sandpiles,"
 * AMERICAN POLITICAL SCIENCE REVIEW 97 (February 2003): 135-150.
 *
 * The Model class constitutes the very core of the model. Here agents are created and put in
 * their territorial context. The main simulation loop can be found in the step() method.
 * The model is a dynamic network model with endogenous geopolitical actors residing in a
 * lattice. The actors' boundaries grow through conquest (see structuralChange()).
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @version 3.0
 */
public class Model extends SimpleModel {
    /**
     * World size along x-axis
     */
    int worldXSize;

    /**
     * World size along z-axis
     */
    int worldYSize;

    /**
     * Number of states created by initial module
     */
    int initPolarity;

    /**
     * Probability of trying unprovoked attack
     */
    double probAttack;

    /**
     * Superiority threshold
     */
    double superiorityThreshold;

    /**
     * Logistical slope
     */
    double superioritySlope;

    /**
     * Proportion of fungible front-allocated resources
     */
    double propMobileResources;

    /**
     * Lowest threshold for invocation of alliance
     */
    double minThreat;

    /**
     * Stopping time for batch runs
     */
    int runUntil;

    /**
     * Main 2D territorial grid
     */
    Object2DGrid world;

    /**
     * Main collection holding actors
     */
    List actors;

    /**
     * Collection holding sovereign actors
     */
    List sovActors;

    /**
     * Counter indicating number of sovereign actors
     */
    int numSovActors;

    /**
     * Counter indicating number of non-atomic actors
     */
    int numCompoundActors;

    /**
     * Counter indicating number of secure actors
     */
    int numSecureActors;

    /**
     * Flag indicating structural change
     */
    private boolean claimEvent;

    /**
     * Method that sets up the default value of the model parameters
     * Inherited from Repast's SimpleModel
     */
    public void setup() {
        super.setup();   // first call setup() in simpleModel

        name = "Geosim";

        // Reset main data structures
        actors = null;
        sovActors = null;
        world = null;

        // Set default parameters
        seed = 1;
        worldXSize = 30;
        worldYSize = 30;
        initPolarity = 200;

        probAttack = 0.2;
        superiorityThreshold = 3.0;
        superioritySlope = 0;
        propMobileResources = 1.0;

        minThreat = -superiorityThreshold;

        // Measurement and display parameters

        numSecureActors = 0;

        runUntil = 10000;

        // Resetting model variables
        claimEvent = false;
    }

    //////// Build Model //////////////////////////////////////////////////////////////////

    /**
     * Method that sets up the model architecture. Called by Repast's SimpleModel.
     */
    public void buildModel() {
        super.buildModel();
        int x, y;

        world = new Object2DGrid(worldXSize, worldYSize);   // create the main territorial grid

        actors = new ArrayList();       // create the main collections
        sovActors = new ArrayList();    // ... this list creates a shortcut to all sovereign actors

        // the 2D grid needs to be populated with agents.
        for (x = 0; x < worldXSize; x++)
            for (y = 0; y < worldYSize; y++) {
                final Actor anActor = new Actor(x * worldYSize + y, // ID
                        x, // position
                        y,
                        this);		// the model
                actors.add(anActor);
                sovActors.add(anActor);
                world.putObjectAt(x, y, anActor);
            }


        // make sure all the agents are shuffled in terms of execution order
        SimUtilities.shuffle(sovActors);

        // create the network of neighbor relations based on the von Neumann neighborhood
        setNeighborhoods();

        numSovActors = sovActors.size();    // set the polarity counter
        numCompoundActors = numSovActors;

        // trigger initial module if switch is true
        initModule();
        makeList();
    }

    ///// Main simulation loop //////////////////////////////////////////////

    /**
     * The main simulation is invoked by SimpleModel.
     */
    public void step() {
        super.step();

        resourceUpdating();

        resourceAllocation();              // let the states allocate resources to their fronts

        decisions();                  // main behavioral phase

        interactions();               // compute the result of interactions

        if (claimEvent) {             // if there's at least one territorial claim...
            structuralChange();         // ... then carry out structural change
            makeList();                 // ... and scramble the execution order
        }
    }

    /**
     * Methods setting up the neighborhood relations used in buildModel().
     */
    private void setNeighborhoods() {
        for (Iterator it1 = actors.iterator(); it1.hasNext();) {
            final Actor actor = (Actor) it1.next();
            final List neighbors = world.getVonNeumannNeighbors(actor.x,  actor.y, false);
            for (Iterator iterator = neighbors.iterator(); iterator.hasNext();) {
                final Actor neighbor = (Actor) iterator.next();
                if(actor.id < neighbor.id) {
                    Relation.createBilateral(actor, neighbor, true, false);
                    actor.numPolNeighs++;
                    neighbor.numPolNeighs++;
                }
            }
        }
    }

    /**
     * Generates log-normal output
     */
    private void initModule() {
        final ArrayList shortList = new ArrayList();
        for (int i = 0; i < initPolarity; i++) {
            final Actor a = (Actor) sovActors.get(i);
            shortList.add(a);
            a.founder = true;
        }

        int j = 0;
        final int n = worldXSize * worldYSize - initPolarity;
        do {
            final int i = Random.uniform.nextIntFromTo(0, shortList.size() - 1);
            final Actor a = (Actor) shortList.get(i);
            if (absorb(a)) {
                shortList.add(a);
                j++;
            }
        } while (j < n);

    }

    /**
     * Method used by initModule() to grow states artificially
     *
     * @param actor
     * @return
     */
    private boolean absorb(final Actor actor) {
        final Relation victim;
        final ArrayList victims = new ArrayList();
        final ListIterator lit = actor.neighs.listIterator();
        for (int i = 0; i < actor.neighs.size(); i++) {
            final Relation r = (Relation) lit.next();
            if (r.pol && r.other.isAtom() && !r.other.founder) {
                victims.add(r);
            }
        }
        if (victims.size() > 0) {
            victim = (Relation) victims.get(Random.uniform.nextIntFromTo(0, victims.size() - 1));
            actor.selectPath(victim);
            conquest(actor, victim.target);
            return true;
        } else
            return false;
    }

    /////////////////////////////////////////////////////////////////////////
    //  Iterated methods /// Called by step() ///////////////////////////////
    /////////////////////////////////////////////////////////////////////////

    //// Resource updating ///////////////////////////////////////////////////

    /**
     * This method constitutes the first phase in the main simulation loop.
     * All actors get to update their resource levels based on territorial
     * resources.
     */
    private void resourceUpdating() {
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            actor.updateResources();
        }
    }

    ////// Resource allocations ////////////////////////////////////////////////////////

    /**
     * The states cannot act without having resources allocated to each of their
     * fronts
     * This method does that for all states that are sovereign
     */
    private void resourceAllocation() {
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            actor.allocateResources();
        }
    }

    ///// Decision phase ///////////////////////////////////////////////////////////////

    /**
     * This phase determines the actors' strategies.
     * The resulting actions are stored in the act field of the Relation objects.
     * The actors have access to the action history from the last move as in a
     * Markov process.
     * Note that double-buffering is used here. The interact() method below updates
     * the buffers.
     */
    private void decisions() {
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            actor.decide();
        }
    }

    ////// Interactions ////////////////////////////////////////////////////////////////

    /**
     * Main interaction method looping through all bi-directional relations once
     * The method is called in step()
     */
    private void interactions() {
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            final ListIterator lit = actor.neighs.listIterator();
            for (int j = 0; j < actor.getNumPoliticalNeighbors(); j++) {
                final Relation r = (Relation) lit.next();
                if (actor.id < r.other.id)
                    interact(r);
            }

        }

    }

    /**
     * In this phase, the states are matched against their political neighbors.
     * The current implementation only allows contiguous territorial interactions.
     * This method carries out a two-way interaction for relation r
     *
     * @param r
     */
    private void interact(final Relation r) {
        // first make sure that two unprovoked attacks do not interfere
        if ((r.isDefective()) && (r.twin.isDefective()) &&
                (r.history[Relation.SELF] == Relation.C) &&
                (r.history[Relation.OTHER] == Relation.C)) {
            if (Random.uniform.nextIntFromTo(0, 1) == 0)
                r.act = Relation.C;
            else
                r.twin.act = Relation.C;
        }

        // the relational objects need to be updated to refresh the history buffers etc
        r.update();
        r.twin.update();

        // main interaction routine that calculates the damage and claims
        // for all four interaction possibilities (C,C), (C,D), (D,C), and (D,D)
        if ((r.isCooperative()) && (r.twin.isCooperative())) {
        } else if ((r.isCooperative()) && (r.twin.isDefective())) {
            r.target = r.twin.agent;
            r.agent = r.twin.target;
            r.twin.offensive = true;
            r.offensive = false;
        } else if ((r.isDefective()) && (r.twin.isCooperative())) {
            r.twin.agent = r.target;
            r.twin.target = r.agent;
            r.offensive = true;
            r.twin.offensive = false;
        } else { // full-fledged war
            // battle routines resulting in claims stored in the relational objects
            // both sides have a theoretical chance to win the battle
            // in the unlikely event of a double claim, this is resolved in structuralChange()

            battle(r);
            battle(r.twin);

            if ((r.claim && !r.offensive) || (r.twin.claim && !r.twin.offensive))
                draw(r);     // reset claims if stalemate
        }
    }

    /**
     * Battle routine for a one-way relation
     *
     * @param r
     */
    private void battle(final Relation r) {
        double victoryThreshold = this.superiorityThreshold;
        final double victorySlope = this.superioritySlope;
        final double ownRes = r.res;
        final double otherRes = r.twin.res;
        if (!r.offensive)
            victoryThreshold = 1.0 / victoryThreshold;
        if (otherRes <= 0.0)
            r.claim = true;
        else if (victorySlope == 0.0)  // no noise
            r.claim = ownRes > otherRes * victoryThreshold;
        else  // this is the logistical criterion based on victoryThreshold, and victC
            r.claim = Utility.bern(Utility.logistic(ownRes / otherRes, victoryThreshold, victorySlope));
        if (r.claim)
            claimEvent = true; // set the claim event flag to invoke structural change
    }

    /**
     * Stalemate method resetting the combat
     *
     * @param r
     */
    private static void draw(final Relation r) {
        r.resetAction();
        r.twin.resetAction();
    }

    ///// Structural change /////////////////////////////////////////////////////

    // In the sctructural-change phase, the states' boundaries are altered
    // as a consequence of conquest.

    /**
     * This method adds an atomic "child" actor to a (possibly compound) "capital" actor
     *
     * @param capital
     * @param province
     */
    private void addProvince(final Actor capital, final Actor province) {
        numSovActors--;

        // create a new relationship between capital and province if necessary
        Relation r = capital.getRelationWith(province, capital.neighs);
        if (r == null)
            r = Relation.createBilateral(capital, province, false, true);
        else {
            r.deactivate();
            r.prov = true;
        }
        r.twin.pol = true;
        capital.numProvinces++;
        province.capital = capital;
        province.numProvinces = 0;

        final double tribute = province.res;
        capital.res = capital.res + tribute;
        province.res = province.res - tribute;


        // update the neighborhood portfolios
        // start by deleting the old relationships of the province
        final ArrayList toBeDel = new ArrayList();
        final Iterator i = province.neighs.iterator();
        while (i.hasNext()) {   // to be on the safe side, it's necessary to create a
            final Relation q = (Relation) i.next();  // a temporary list of items to be deleted
            if (q.other != capital) {
                toBeDel.add(q);
            }
        }

        for (int j = 0; j < toBeDel.size(); j++) {
            final Relation q = (Relation) toBeDel.get(j);
            q.other.neighs.remove(q.twin);
            q.owner.neighs.remove(q);
        }

        // finally, figure out if the new capital "inherits" political relations from province
        // note that the processing is over vonNeumann neighbors
        final List neighbors = province.getNeighbors();
        for (Iterator iterator = neighbors.iterator(); iterator.hasNext();) {
            final Actor neighbor = (Actor) iterator.next();
            if(province.isForeigner(neighbor)) {
                // The following method makes sure that you don't already
                // have the same relation
                Relation.createIfNotExistBilateral(capital,
                        neighbor.getGovt(),
                        true,
                        false);
            }
        }
    }

    /**
     * Method that removes an atomic "child" actor from a compound capital actor
     *
     * @param capital
     * @param province
     */
    private void removeProvince(final Actor capital, final Actor province) {
        resetFronts(province);
        numSovActors++;
        capital.numProvinces--;
        province.capital = null;

        // adjust the neighbor portfolios
        final Relation r = capital.getRelationWith(province, capital.neighs);
        if (r != null) {
            // start by severing the "parental" bond
            r.resetAction();
            r.twin.resetAction();

            capital.neighs.remove(r);
            province.neighs.remove(r.twin);

            // There is an assumption here that no other relation exists for the child
            // than the one to its capital

            // then go through all four territorial neighbors of the newly independent "child"
            final List neighbors = province.getNeighbors();
            for (Iterator iterator = neighbors.iterator(); iterator.hasNext();) {
                final Actor neighbor = (Actor) iterator.next();
                final Relation newRel = Relation.createIfNotExistBilateral(province,
                        neighbor.getGovt(), true, false);
                newRel.resetAction();
                newRel.twin.resetAction();
                final Relation oldCapitalRel = capital.getRelationWith(neighbor.getGovt(), capital.neighs);
                if (oldCapitalRel != null)
                    if ((!neighbor.getGovt().isStateAdjacentTo(capital))) {
                        neighbor.getGovt().neighs.remove(oldCapitalRel.twin);
                        capital.neighs.remove(oldCapitalRel);
                    }

            }
        } else {// this should never happen
            System.out.println("Child has no capital error!");
        }

        final ListIterator lit = province.neighs.listIterator();
        for (int j = 0; j < province.getNumPoliticalNeighbors(); j++) {
            final Relation rel = (Relation) lit.next();
            rel.res = (propMobileResources * province.res) +
                    (((1 - propMobileResources) * province.res) /
                    province.getNumPoliticalNeighbors());
        }
    }

    /**
     * Method that strips an actor of all of its provinces
     * @param actor
     */
    private void collapse(final Actor actor) {
        final ArrayList toBeDel = new ArrayList();

        final Iterator lit = actor.neighs.iterator();
        while (lit.hasNext()) {
            final Relation r = (Relation) lit.next();
            if (r.prov)
                toBeDel.add(r.other);
        }

        for (int j = 0; j < toBeDel.size(); j++)
            removeProvince(actor, (Actor) toBeDel.get(j));
    }

    /**
     * Method used in topology() below
     * The method marks provinces to check contiguity
     *
     * @param current
     */
    private void reach(final Actor current) {
        current.reached = true;
        final List neighbors = current.getNeighbors();
        for (Iterator iterator = neighbors.iterator(); iterator.hasNext();) {
            final Actor neighbor = (Actor) iterator.next();
            if (current.isCompatriot(neighbor) && !neighbor.reached)
                reach(neighbor); // recursive call

        }
    }

    /**
     * Method that guarantees territorial contiguity within compound actors.
     * All unconnected provinces that cannot be reached by capital
     * are removed from the state
     *
     * @param actor
     */
    private void topology(final Actor actor) {
        // Reset reach
        actor.reached = true;
        ListIterator lit = actor.neighs.listIterator();
        for (int i = 0; i < actor.neighs.size(); i++) {
            final Relation r = (Relation) lit.next();
            if (r.prov)
                r.other.reached = false;
        }

        reach(actor);

        // remove the ones cut off
        final ArrayList toBeDel = new ArrayList();

        lit = actor.neighs.listIterator();
        while (lit.hasNext()) {
            final Relation r = (Relation) lit.next();
            if ((r.prov) && (!r.other.reached))
                toBeDel.add(r.other);
        }

        for (int j = 0; j < toBeDel.size(); j++) {
            final Actor a = (Actor) toBeDel.get(j);
            removeProvince(actor, a);
        }
    }


    private static void resetFronts(final Actor target) {
        final Actor govt = target.getGovt();
        final ListIterator lit = govt.neighs.listIterator();
        for (int i = 0; i < govt.neighs.size(); i++) {
            final Relation r = (Relation) lit.next();
            if (r.pol && (r.agent == target)) {
                r.resetAction();
                r.twin.resetAction();
            }
        }
    }

    /**
     * Method carrying out conquest by actor of target
     * NOTE: the target is the territorial unit to be incorporated and
     * not the political actor losing the territory
     * @param actor
     * @param target
     */
    private void conquest(final Actor actor, final Actor target) {
        resetFronts(target);
        if (!target.isSovereign()) { // if the target is a province in another state...
            final Actor invaded = target.capital;
            removeProvince(invaded, target);
            if (!invaded.isAtom())  // need to check territorial contiguity inside losing state
                topology(invaded);
        } else if (!target.isAtom()) // if the target is a capital province...
            collapse(target);          // ... then collapse entire state
        addProvince(actor, target);
    }

    /**
     * Method that carries out structural change
     * in the current simple version of the model, this amounts to conquest
     */
    private void structuralChange() {
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            if (actor.isSovereign()) {
                for (int j = 0; j < actor.neighs.size(); j++) {  // ... and their political neighbors
                    final Relation r = (Relation) actor.neighs.get(j);
                    if (r.pol && r.other.isSovereign()) {
                        if (r.claim)          // ... checking for claims making sure no previous territorial change
                            if ((r.agent.getGovt() == actor) && (r.target.getGovt() == r.other)) { // confounds operation
                                r.claim = false;
                                r.twin.claim = false;
                                final Actor temp = r.target;
                                r.resetAction();
                                r.twin.resetAction();
                                conquest(actor, temp);
                            } else {
                                r.resetAction();
                                r.twin.resetAction();
                            }
                    }
                }
            }
        }
    }

    /**
     * Method that updates and shuffles the list of sovereign actors
     * in case of structural change
     */
    private void makeList() {
        claimEvent = false;
        numCompoundActors = 0;
        sovActors.clear();

        for (Iterator iterator = actors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            actor.numPolNeighs = actor.neighs.size() - actor.numProvinces;
            if (actor.isSovereign()) {
                sovActors.add(actor);
                if (!actor.isAtom())
                    numCompoundActors++;
            }
        }
        SimUtilities.shuffle(sovActors);    // scramble the list to randomize execution order
    }

    protected void perceptions() {
        numSecureActors = 0;
        for (Iterator iterator = sovActors.iterator(); iterator.hasNext();) {
            final Actor actor = (Actor) iterator.next();
            actor.perceive();

        }
    }

    /**
     * Creating and starting your model
     * @param args
     */
    public static void main(final String[] args) {
        final SimInit init = new SimInit();
        final Model m = new Model();
        init.loadModel(m, null, false);
    }
}

