/*
 * Copyright (c) 2004 Center for Comparative and International Studies
 * at the Swiss Federal Institute of Technology Zurich
 * (see http://www.cis.ethz.ch/ for details)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Please see http://www.gnu.org/copyleft/gpl.txt for the full license text.
 *
 */
package ch.ethz.icr.geosim0;

import uchicago.src.sim.analysis.LocalDataRecorder;
import uchicago.src.sim.engine.Controller;
import uchicago.src.sim.engine.SimInit;

/**
 * <p>Title: Geosim: Geopolitical Simulation Model</p>
 * <p>Description: Model that simulates state formation with endogenous boundaries</p>
 * <p>Copyright: Copyright (c) March 11, 2002</p>
 * <p>Company: Harvard University</p>
 *
 * @author Lars-Erik Cederman
 * @version 2.0
 */

/**
 * Geosim: Geopolitical Simulation Model
 * A model to simulate state formation with endogenous boundaries
 *
 * This model is based on the architecture introduced in Cederman, EMERGENT ACTORS IN WORLD
 * POLITICS. Princeton University Press, 1997. The current model corresponds to the more
 * recent article Cederman, "Modeling the Size of Wars: From Billiard Balls to Sandpiles,"
 * AMERICAN POLITICAL SCIENCE REVIEW 97 (February 2003): 135-150.
 *
 * Batch interface of the model implemented as an extension of the Model.
 * The class customizes the parameter reader and prepares the file output.
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @version 3.0
 */
public class ModelBatch extends Model {
    /**
     * Repast's data recorder
     */
    private LocalDataRecorder recorder;

    public ModelBatch() {
        // Setting the order of appearance of parameters
        // to NON-alphabetic
        Controller.ALPHA_ORDER = false;
    }

    /**
     * Method setting up the simulation
     */
    public final void setup() {
        super.setup();   // first call setup() in Model
        System.gc();
        params = new String[]{"seed", "runUntil", "worldXSize", "worldYSize",
                              "probAttack", "superiorityThreshold", "superioritySlope",
                              "propMobileResources"};
    }

    /**
     * Method creating the instrumentation to record data to file
     */
    public final void buildModel() {
        super.buildModel();   // first call buildModel() in Model
        setStoppingTime(runUntil);

        recorder = new LocalDataRecorder("./data.txt", this);
        recorder.setDelimiter(" ");

        recorder.createNumericDataSource("sov", this, "computePolarity", 4, 3);
        recorder.createNumericDataSource("sec",this,  "computeSecurity", 2, 3);
    }

    /**
     * Method carried out every time step
     */
    public final void step() {
        super.step();      // first call step() in Model

        perceptions();

        if ((getTickCount() == runUntil) || (getTickCount() % 1000 == 0.0) || (getTickCount() == 1)) {
            System.out.println(getTickCount() + " Sov actors =" + numSovActors + " / " + numCompoundActors + " " + (double)numSecureActors / (double)(worldXSize * worldYSize));
        }
        if ((getTickCount() == runUntil) /*|| (getTickCount() % 500 == 0.0)*/) {
            recorder.record();

        }
    }

    public final void atEnd() {
        super.atEnd();
        recorder.write();
    }

    public final int getRunUntil() {
        return runUntil;
    }

    public final void setRunUntil(final int v) {
        runUntil = v;
    }

    public int getWorldXSize() {
        return worldXSize;
    }

    public void setWorldXSize(final int worldXSize) {
        this.worldXSize = worldXSize;
    }

    public int getWorldYSize() {
        return worldYSize;
    }

    public void setWorldYSize(final int worldYSize) {
        this.worldYSize = worldYSize;
    }

    public int getInitPolarity() {
        return initPolarity;
    }

    public void setInitPolarity(final int initPolarity) {
        this.initPolarity = initPolarity;
    }

    public double getProbAttack() {
        return probAttack;
    }

    public void setProbAttack(final double probAttack) {
        this.probAttack = probAttack;
    }

    public double getSuperiorityThreshold() {
        return superiorityThreshold;
    }

    public void setSuperiorityThreshold(final double superiorityThreshold) {
        this.superiorityThreshold = superiorityThreshold;
    }

    public double getSuperioritySlope() {
        return superioritySlope;
    }

    public void setSuperioritySlope(final double superioritySlope) {
        this.superioritySlope = superioritySlope;
    }

    public double getPropMobileResources() {
        return propMobileResources;
    }

    public void setPropMobileResources(final double propMobileResources) {
        this.propMobileResources = propMobileResources;
    }

    //////////////////////////////////////////////////////////////

    public final Integer computeSov() {
        return new Integer(numSovActors);
    }

    public final double computePolarity() {
        return (double)numSovActors;
    }

    public final double computeSecurity() {
        return (double) numSecureActors / (double)(worldXSize * worldYSize);
    }

    /**
     * Creating and starting your model
     *
     * @param args
     */
    public static void main(final String[] args) {
        final SimInit init = new SimInit();
        final ModelBatch m = new ModelBatch();
        init.loadModel(m, "./params.txt", true);
    }

}


