/*
 * Copyright (c) 2004 Center for Comparative and International Studies
 * at the Swiss Federal Institute of Technology Zurich
 * (see http://www.cis.ethz.ch/ for details)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Please see http://www.gnu.org/copyleft/gpl.txt for the full license text.
 *
 */
package ch.ethz.icr.geosim0;

import uchicago.src.sim.analysis.OpenSequenceGraph;
import uchicago.src.sim.analysis.Sequence;
import uchicago.src.sim.engine.Controller;
import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.gui.DisplayConstants;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.gui.Object2DDisplay;

import java.util.Iterator;

/**
 * Geosim: Geopolitical Simulation Model
 * A model to simulate state formation with endogenous boundaries
 *
 * This model is based on the architecture introduced in Cederman, EMERGENT ACTORS IN WORLD
 * POLITICS. Princeton University Press, 1997. The current model corresponds to the more
 * recent article Cederman, "Modeling the Size of Wars: From Billiard Balls to Sandpiles,"
 * AMERICAN POLITICAL SCIENCE REVIEW 97 (February 2003): 135-150.
 *
 * The ModelGUI class creates a graphical interface for the Model featuring
 * a dynamic graph for plotting polarity and security.
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @version 3.0
 */
public final class ModelGUI extends Model {
    /**
     * Main display surface
     */
    private DisplaySurface dsurf;

    /**
     *  The polarity graph
     */
    private OpenSequenceGraph polarityGraph;

    /**
     * The security graph
     */
    private OpenSequenceGraph securityGraph;

    /**
     * Switch for polarity graph
     */
    private boolean showPolarityGraph;

    /**
     * Switch for security graph
     */
    private boolean showSecurityGraph;

    /**
     * The constructor of the GUI
     */
    public ModelGUI() {
        // Setting the order of appearance of parameters
        // to NON-alphabetic
        Controller.ALPHA_ORDER = false;
        Controller.CONSOLE_ERR = false;
        Controller.CONSOLE_OUT = false;
    }

    final class GraphSeq implements Sequence {
        public final double getSValue() {
            return (double) numSovActors;
        }
    }

     final class GraphSeqPropSecurity implements Sequence {
         public final double getSValue() {
             return (double) numSecureActors / (double)(worldXSize * worldYSize);
         }
     }

    /**
     * Method setting up the simulation
     */
    public final void setup() {

        super.setup();   // call to the setup() method of Model

        params = new String[]{"seed", "worldXSize", "worldYSize",
                              "probAttack", "superiorityThreshold", "superioritySlope",
                              "propMobileResources",
                              "ShowPolarityGraph", "ShowSecurityGraph"};

        if (dsurf != null)  // remove the old display, if new run...
            dsurf.dispose();
        dsurf = null;

        if (polarityGraph != null)  // remove old power plot, if new run...
            polarityGraph.dispose();
        if (securityGraph != null)  // remove old power plot, if new run...
            securityGraph.dispose();
        System.gc();

        dsurf = new DisplaySurface(this, "Geosim Display");
        registerDisplaySurface("Main", dsurf);

        DisplayConstants.CELL_WIDTH = 20;
        DisplayConstants.CELL_HEIGHT = 20;

        showPolarityGraph = true;
        showSecurityGraph = true;

        System.out.println("GeoSim v. 0");
        System.out.println();
    }

    /**
     * Method that sets up the instrumentation for the GUI
     */
    public final void buildModel() {
        super.buildModel();   // call to buildModel() in Model

        // create display surface
        final Object2DDisplay display = new Object2DDisplay(world);
        display.setObjectList(actors);
        dsurf.addDisplayableProbeable(display, "Display");
        addSimEventListener(dsurf);
        dsurf.display();

        // create graph showing the number of states
        if (showPolarityGraph) {
            polarityGraph = new OpenSequenceGraph("Polarity", this);
            polarityGraph.setXRange(0, 50);
            polarityGraph.setYRange(0.0, 1.0);
            polarityGraph.setAxisTitles("Time", "Number");
            polarityGraph.addSequence("Sov", new GraphSeq());
            polarityGraph.display();
            polarityGraph.step();
        }

        if (showSecurityGraph) {
            securityGraph = new OpenSequenceGraph("Security", this);
            securityGraph.setXRange(0, 50);
            securityGraph.setYRange(0.0, 1.1);
            securityGraph.setAxisTitles("Time", "Number");
            securityGraph.addSequence("Secure", new GraphSeqPropSecurity());
            securityGraph.display();
        }
    }

    /**
     * Recurrent method called each time step
     */
    public final void step() {

        super.step();  // first execute the step() method of Model

        perceptions();

        if (showPolarityGraph)
            polarityGraph.step();
        if (showSecurityGraph)
            securityGraph.step();

        dsurf.updateDisplay();  // update the display surface
    }

    public int getWorldXSize() {
        return worldXSize;
    }

    public void setWorldXSize(final int worldXSize) {
        this.worldXSize = worldXSize;
    }

    public int getWorldYSize() {
        return worldYSize;
    }

    public void setWorldYSize(final int worldYSize) {
        this.worldYSize = worldYSize;
    }

    public int getInitPolarity() {
        return initPolarity;
    }

    public void setInitPolarity(final int initPolarity) {
        this.initPolarity = initPolarity;
    }

    public double getProbAttack() {
        return probAttack;
    }

    public void setProbAttack(final double probAttack) {
        this.probAttack = probAttack;
    }

    public double getSuperiorityThreshold() {
        return superiorityThreshold;
    }

    public void setSuperiorityThreshold(final double superiorityThreshold) {
        this.superiorityThreshold = superiorityThreshold;
    }

    public double getSuperioritySlope() {
        return superioritySlope;
    }

    public void setSuperioritySlope(final double superioritySlope) {
        this.superioritySlope = superioritySlope;
    }

    public double getPropMobileResources() {
        return propMobileResources;
    }

    public void setPropMobileResources(final double propMobileResources) {
        this.propMobileResources = propMobileResources;
    }

    public final boolean getShowPolarityGraph() {
        return showPolarityGraph;
    }

    public final void setShowPolarityGraph(final boolean showPolarityGraph) {
        this.showPolarityGraph = showPolarityGraph;
    }

    public final boolean isShowSecurityGraph() {
        return showSecurityGraph;
    }

    public final void setShowSecurityGraph(final boolean showSecurityGraph) {
        this.showSecurityGraph = showSecurityGraph;
    }

    /**
     * Creating and starting your model
     * @param args
     */
    public static void main(final String[] args) {
        final SimInit init = new SimInit();
        // We MUST create a ModelGUI object instead of an instance of Model
        // as in the capital class, in order to get the GUI functionality.
        final Model m = new ModelGUI();
        init.loadModel(m, null, false);
    }
}