/*
 * Copyright (c) 2004 Center for Comparative and International Studies
 * at the Swiss Federal Institute of Technology Zurich
 * (see http://www.cis.ethz.ch/ for details)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Please see http://www.gnu.org/copyleft/gpl.txt for the full license text.
 *
 */
package ch.ethz.icr.geosim0;

import java.util.ListIterator;

/**
 * Geosim: Geopolitical Simulation Model
 * A model to simulate state formation with endogenous boundaries
 *
 * This model is based on the architecture introduced in Cederman, EMERGENT ACTORS IN WORLD
 * POLITICS. Princeton University Press, 1997. The current model corresponds to the more
 * recent article Cederman, "Modeling the Size of Wars: From Billiard Balls to Sandpiles,"
 * AMERICAN POLITICAL SCIENCE REVIEW 97 (February 2003): 135-150.
 *
 * The Relation class defines the network edges used to build up neighbor lists.
 * All relations are bi-directional, which means that between states i and j
 * there's always a relation i to j and one from j to i.
 * The edges hold important information pertaining to all types of relations.
 * Political relations point to all strategic neighbors.
 * Parent-child relations relates to hierarchical relations between a capital
 * and dependent provinces.
 * NOTE: To speed up execution, political relations are always put at the beginning
 * of the neighbor lists.
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @version 3.0
 */
public final class Relation {
    /**
     * The first node of the relationship
     */
    final Actor owner;

    /**
     * The second node of the relationship
     */
    final Actor other;

    /**
     * A pointer to the relation in the other direction
     */
    Relation twin;

    /**
     * Relational resources and damage
     */
    double res, oldRes;

    /**
     * Action decided by owner (C or D)
     */
    int act;                    //

    /**
     * The units of the battle path
     */
    Actor agent, target;

    /**
     * The strategic buffer
     */
    int[] history = new int[2];

    /**
     * Type of relationship
     */
    boolean pol, prov;

    /**
     * Claim flag indicating intention to conquer
     */
    boolean claim;

    /**
     * Flag used to indicate direction of attack
     */
    boolean offensive;

    //// Constants used to index history[] /////////////////////////////

    public static final int SELF = 0;
    public static final int OTHER = 1;
    public static final int C = 0;  // Conflict (used in the act record of Relation)
    public static final int D = 1;  // Defect

    ////////////////////////////////////////////////////////////////////

    /**
     * Method that creates a bi-directional relationship from scratch
     *
     * @param a1
     * @param a2
     * @param isPolitical
     * @param isChild
     * @return
     */
    public static Relation createBilateral(final Actor a1, final Actor a2,
                                           final boolean isPolitical,
                                           final boolean isChild) {

        final Relation r1 = new Relation(a1, a2, isPolitical, isChild);
        final Relation r2 = new Relation(a2, a1, isPolitical, false);
        r1.twin = r2;
        r2.twin = r1;
        if (isPolitical) {  // make sure that political relationships go first
            a1.neighs.addFirst(r1);
            a2.neighs.addFirst(r2);
        } else {
            a1.neighs.addLast(r1);
            a2.neighs.addLast(r2);
        }

        return r1;
    }

    /**
     * Method that creates a new bi-directional relationship
     *
     * @param a1
     * @param a2
     * @param isPolitical
     * @param isChild
     * @return
     */
    public static Relation createIfNotExistBilateral(final Actor a1, final Actor a2,
                                                     final boolean isPolitical,
                                                     final boolean isChild) {
        final ListIterator lit = a1.neighs.listIterator();
        for (int i = 0; i < a1.neighs.size(); i++) {
            final Relation r = (Relation) lit.next();
            if ((r.owner == a1) && (r.other == a2)) {
                if (isPolitical) {
                    r.activate();
                } else {
                    r.deactivate();
                }
                return r;
            }
        }

        return createBilateral(a1, a2, isPolitical, isChild);
    }

    /**
     * The constructor
     *
     * @param owner
     * @param other
     * @param isPolitical
     * @param isChild
     */
    private Relation(final Actor owner, final Actor other,
                    final boolean isPolitical, final boolean isChild) {

        this.owner = owner;
        this.other = other;
        pol = isPolitical;
        prov = isChild;
        claim = false;
        resetAction();
    }


    /**
     * Method that resets a conflict including resetting the history buffers
     */
    public final void resetAction() {
        res = 0.0;
        act = C;
        agent = null;
        target = null;
        history = new int[2];
        history[SELF] = C;
        history[OTHER] = C;
        claim = false;
        offensive = false;
    }

    /**
     * Update the history buffer
     */
    public final void update() {
        history[SELF] = act;
        history[OTHER] = twin.act;
    }

    /**
     * Method that puts the political neighbors at the head of the neighbor list
     * this speeds up execution
     */
    private void activate() {
        pol = true;
        twin.pol = true;
        owner.neighs.remove(this);
        owner.neighs.addFirst(this);
        other.neighs.remove(twin);
        other.neighs.addFirst(twin);
        resetAction();
        twin.resetAction();
    }

    /**
     * Method that puts inactive neighbors at the end of the neighbor list
     */
    public final void deactivate() {
        pol = false;
        twin.pol = false;
        owner.neighs.remove(this);
        owner.neighs.addLast(this);
        other.neighs.remove(twin);
        other.neighs.addLast(twin);
        resetAction();
        twin.resetAction();
    }

    public final boolean isCooperative() {
        return act == Relation.C;
    }

    public final boolean isDefective() {
        return act == Relation.D;
    }
}
